%%% -*- erlang-indent-level: 2 -*-
%%%-------------------------------------------------------------------
%%% File    : dialyzer.erl
%%% Authors : Tobias Lindahl <tobiasl@csd.uu.se>
%%%           Kostis Sagonas <kostis@it.uu.se>
%%% Description : The interface for the Dialyzer tool.
%%%
%%% Created : 27 Apr 2004 by Tobias Lindahl <tobiasl@csd.uu.se>
%%%-------------------------------------------------------------------
-module(dialyzer).

-export([start/3]).

-include("dialyzer.hrl").	  %% file is automatically generated
-include("hipe_icode_type.hrl").

-record(state, {parent, legal_warnings}).

start(Parent, LegalWarnings, Analysis) ->
  State = #state{parent=Parent, legal_warnings=LegalWarnings},
  NewAnalysis1 = expand_files(Analysis),
  init_plt(Analysis),
  NewAnalysis2 = run_analysis(NewAnalysis1),
  loop(State, NewAnalysis2).

expand_files(Analysis) ->
  Files = Analysis#analysis.files,
  Ext = 
    case Analysis#analysis.start_from of
      byte_code -> ".beam";
      src_code -> ".erl"
    end,
  case expand_files(Files, Ext, []) of
    [] ->
      exit({error, "No files to analyse. Check analysis type."});
    NewFiles ->
      Analysis#analysis{files=NewFiles}
  end.

expand_files([File|Left], Ext, Acc) ->
  case filelib:is_dir(File) of
    true ->
      {ok, List} = file:list_dir(File),
      NewFiles = [filename:join(File, X) 
		  || X <- List, filename:extension(X)==Ext],
      expand_files(Left, Ext, NewFiles++Acc);
    false ->
      expand_files(Left, Ext, [File|Acc])
  end;
expand_files([], _Ext, Acc) ->
  ordsets:from_list(Acc).

%% ----------------------------------------------------------------
%%
%%  Main Loop
%%


loop(State, Analysis = #analysis{}) ->
  AnalPid = Analysis#analysis.analysis_pid,
  Parent = State#state.parent,

  receive
    {AnalPid, log, LogMsg} ->
      send_log(Parent, LogMsg),
      loop(State, Analysis);
    {AnalPid, warnings, Warnings} ->
      case filter_warnings(State#state.legal_warnings, Warnings) of
	[] -> ok;
	SendWarnings ->
	  send_warnings(Parent, SendWarnings)
	end,
      loop(State, Analysis);
    {AnalPid, error, Msg} ->
      send_error(Parent, Msg),
      loop(State, Analysis);
    {AnalPid, module_callgraph, MCG} ->
      loop(State, Analysis#analysis{mcg={module_callgraph, MCG}});
    {AnalPid, done} ->
      NewAnalysis = analysis_done(Parent, Analysis),
      loop(State, NewAnalysis);
    {Parent, stop} ->
      exit(AnalPid, kill),
      ok;
    _Other ->
      %io:format("Received ~p\n", [Other]),
      loop(State, Analysis)
  end.

%% ----------------------------------------------------------------
%%
%%  Handle messages
%%

send_warnings(Parent, Warnings) ->
  Parent ! {self(), warnings, Warnings}.


filter_warnings(LegalWarnings, Warnings) ->
  [String || {Tag, String} <- Warnings, lists:member(Tag, LegalWarnings)].


send_error(Parent, Msg) ->
  Parent ! {self(), error, Msg}.
  

send_log(Parent, LogMsg) ->
  Parent ! {self(), log, LogMsg}.

send_analysis_done(Parent) ->
  Parent ! {self(), done}.

%% ---- Plt ----

init_plt(#analysis{init_plt=From, user_plt=To, md5=MD5}) ->
  case file:copy(From, To) of
    {ok, _} ->
      case MD5 of
	none -> ok;
	_ ->
	  {ok, Dets} = dets:open_file(To, []),
	  dets:insert(Dets, {md5, MD5}),
	  ok = dets:close(Dets),
	  ok
      end;
    {error, Reason} ->
      io:format("from: ~s\n",[From]),      
      Msg = io_lib:format("Could not init lookup table:\n\n~w",[Reason]),
      exit({error, Msg})
  end.

%% ----------------------------------------------------------------
%%
%%  Run the analysis
%%


run_analysis(Analysis) ->
  Self = self(),
  Fun = 
    fun()-> dialyzer_analysis:start(Self, Analysis)end,
  Analysis#analysis{analysis_pid=spawn_link(Fun)}.

analysis_done(Parent, Analysis) ->
  case Analysis#analysis.fixpoint of
    first ->
      Msg = "\n========== Starting final run ==========\n",
      send_log(Parent, Msg),
      run_analysis(Analysis#analysis{fixpoint=second});
    second ->
      send_analysis_done(Parent),
      exit(normal);
    false ->
      send_analysis_done(Parent),
      exit(normal)
  end.
