%%%-------------------------------------------------------------------
%%% File    : dialyzer_analysis.erl
%%% Author  : Tobias Lindahl <tobiasl@csd.uu.se>
%%% Description : Front end for the analysis employed by the dialyzer.
%%%
%%% Created : 7 May 2004 by Tobias Lindahl <tobiasl@csd.uu.se>
%%%-------------------------------------------------------------------

-module(dialyzer_analysis).
-export([start/2]).

-include("dialyzer.hrl").

-define(OPTS(Pid_), 
	[{type_warnings, {pid, Pid_}}, type_only, no_inline_fp,
	 {use_callgraph, fixpoint}, get_called_modules, 
	 {pmatch, no_duplicates}]).

-record(analysis_state, 
	{  
	  defines,
	  dets,
	  doc_ets,
	  fixpoint,
	  granularity,
	  include_dirs,
	  mcg=none,
	  opts,
	  parent,
	  start_from
	 }).

start(Parent, Analysis) ->
      
  Opts1 = ?OPTS(self()),
  Opts2 =
    case Analysis#analysis.fixpoint of
      false ->
	[use_callgraph | Opts1];
      first ->
	[no_type_warnings|Opts1];
      second ->
	Opts1
    end,
  Opts3 =
    case Analysis#analysis.check_inline of
      true -> [check_for_inlining |Opts2];
      false -> Opts2
    end,

  Opts4 =
    case Analysis#analysis.granularity of
      module -> Opts3;
      all -> [{icode_type, {dets, Analysis#analysis.user_plt}}|Opts3]    
    end,

  State = #analysis_state{defines=Analysis#analysis.defines,
			  dets=Analysis#analysis.user_plt,
			  doc_ets=Analysis#analysis.doc_plt,
			  fixpoint=Analysis#analysis.fixpoint,
			  granularity=Analysis#analysis.granularity,
			  include_dirs=Analysis#analysis.include_dirs,
			  mcg=Analysis#analysis.mcg,
			  opts=Opts4,
			  parent=Parent,
			  start_from=Analysis#analysis.start_from
			 },

  case State#analysis_state.mcg of
    none ->
      analyze_all_files(Analysis#analysis.files, State);
    {module_callgraph, _MCG} ->
      analyze_callgraph(State)
  end,
  done(Parent).


%%============ Analyze files according to a callgraph ==============
analyze_callgraph(State=#analysis_state{mcg ={module_callgraph, Callgraph}}) ->
  analyze_sccs(Callgraph, State).

analyze_sccs(Callgraph, State) ->
  case dialyzer_module_callgraph:is_empty(Callgraph) of
    true ->
      ok;
    false ->
      case dialyzer_module_callgraph:take_first(Callgraph) of
	{[Mod], NewCallgraph} ->
	  %% Only one module
	  analyze_mod(Mod, State),
	  analyze_sccs(NewCallgraph, State);
	{Scc, NewCallgraph} ->
	  Parent = State#analysis_state.parent,
	  send_log(Parent, 
		   "-- Analyzing strongly connected component\n"
		   "-- (holding warnings)\n"),
	  Opts = State#analysis_state.opts,
	  TmpOpts = [no_type_warnings|Opts],
	  TmpState = State#analysis_state{opts=TmpOpts},
	  analyze_one_scc(Scc, TmpState),
	  send_log(Parent,
		   "-- Reached fixpoint for strongly connected component\n"
		   "-- Re-analyzing to get warnings\n"),
	  [analyze_mod(X, State) || X <- Scc],
	  analyze_sccs(NewCallgraph, State)
      end
  end.

analyze_one_scc(Scc, State) ->
  case analyze_one_scc(Scc, State, true) of
    fixpoint -> 
      ok;
    not_fixpoint -> 
      Msg = "-- Re-analyzing strongly connected component --\n",
      send_log(State#analysis_state.parent, Msg),
      analyze_one_scc(Scc, State)
  end.

analyze_one_scc([Mod|Left], State = #analysis_state{dets=Dets}, Fixpoint) ->
  Before = get_module_info(Dets, Mod),
  analyze_mod(Mod, State),
  After = get_module_info(Dets, Mod),
  case info_is_equal(Before, After) of
    true -> analyze_one_scc(Left, State, Fixpoint);
    false -> analyze_one_scc(Left, State, false)
  end;
analyze_one_scc([], _State, Fixpoint) ->
  case Fixpoint of
    true -> fixpoint;
    false -> not_fixpoint
  end.

info_is_equal([{MFA, Type1}|Left1], [{MFA, Type2}|Left2]) ->
  case erl_types:t_is_equal(Type1, Type2) of
    true ->
      info_is_equal(Left1, Left2);
    false ->
      false
  end;
info_is_equal([], []) ->
  true;
info_is_equal(_, _) ->
  false.


%%============ Analyze all files ==============

analyze_all_files(Files, State) ->
  ModMap = get_modules_by_dir(Files, State#analysis_state.start_from),
  case analyze_all_files(ModMap, State, []) of
    ok ->
      ok;
    {module_callgraph, NewMCG} ->
      send_module_callgraph(State#analysis_state.parent, NewMCG)
  end,
  ok.

analyze_all_files([File|Left], State, Acc) ->
  case analyze_mod(File, State) of
    error ->
      analyze_all_files(Left, State, Acc);
    ok ->
      analyze_all_files(Left, State, Acc);
    {called_modules, Called} ->
      analyze_all_files(Left, State, [{File, Called}|Acc])
  end;
analyze_all_files([], _State, Acc) ->
  case Acc of
    [] ->
      ok;
    List when is_list(List) ->
      {module_callgraph, dialyzer_module_callgraph:construct(List)}
  end.

%%============ Analyze one module ==============

analyze_mod(Mod, State) ->
  Res =
    case State#analysis_state.granularity of
      module ->
	TmpDets = mk_tmp_dets(State#analysis_state.dets),
	TmpOpts = [{icode_type, {dets, TmpDets}}|State#analysis_state.opts],
	TmpState = State#analysis_state{dets=TmpDets, opts=TmpOpts},
	analyze_mod_1(Mod, TmpState);
      all ->
	analyze_mod_1(Mod, State)
    end,
  case Res of
    error -> 
      flush(),
      error;
    _ -> 
      insert_doc_info(Mod, State),
      Res
  end.

analyze_mod_1({Mod, Dir}, State = #analysis_state{parent=Parent}) ->
  {ok, CWD} = file:get_cwd(),
  case file:set_cwd(Dir) of
    {error, _} ->
      Msg =io_lib:format("*****Skipping ~s. Could not cd into ~s\n",[Mod, Dir]),
      send_error(Parent, Msg),
      error;
    ok ->
      case analyze_mod_2(Mod, State) of
	{error, ErrorReport} ->
	  send_error(Parent, ErrorReport),
	  ok = file:set_cwd(CWD),
	  error;
	{ok, Msg} ->
	  send_all_warnings(Parent),
	  send_log(Parent, Msg),
	  ok = file:set_cwd(CWD),
	  ok;
	{ok, Msg, CalledModules} ->
	  send_all_warnings(Parent),
	  send_log(Parent, Msg),
	  ok = file:set_cwd(CWD),
	  {called_modules, CalledModules}
      end
  end.

analyze_mod_2(Mod, State = #analysis_state{opts=Opts}) ->
  send_log(State#analysis_state.parent, "Analyzing " ++ Mod),
  T0 = time_now(),
  Res =
    case State#analysis_state.start_from of
      src_code ->
	CompOpts = [no_inline, {i, ".."}, {i, "../incl"}, {i, "../inc"},
		    {i, "../include"},
		    return_errors, 
		    binary, to_core, {core_transform, cerl_typean}],
	Includes = [{i, X} ||X <- State#analysis_state.include_dirs],
	Defines = [{d, M, V} || {M,V} <- State#analysis_state.defines],
	case compile:file(Mod, CompOpts++Includes++Defines) of
	  {ok, _, Core} ->
	    catch hipe:compile_core(list_to_atom(Mod), Core, [], Opts);
	  {error, Errors, _} ->
	    FormatedErrors = format_errors(Errors),
	    {error, lists:flatten(FormatedErrors)}
	end;
      byte_code ->
	catch hipe:c(list_to_atom(Mod), Opts)
    end,
  case Res of
    {{ok, _}, {called_modules, C}} ->
      T = time_since(T0) / 1000,
      {ok, io_lib:format("...done in ~.2f secs\n",[T]), C};
    {ok, _} ->
      T = time_since(T0) / 1000,
      {ok, io_lib:format("...done in ~.2f secs\n",[T])};
    {error, What}  ->
      {error, 
       io_lib:format("\n*****In: ~s\n"
		     "*****Analysis failed with error report:\n~s\n",
		     [Mod, What])};
    {'EXIT', Why} ->
      {error, 
       io_lib:format("\n*****In: ~s\n"
		     "*****Analysis failed with error report:\n~w\n",
		     [Mod, Why])}
  end.

format_errors([{Mod, Errors}|Left])->
  FormatedError = 
    [io_lib:format("~s:~w: ~s\n", [Mod, Line,apply(M,format_error, [Desc])])
     || {Line, M, Desc} <- Errors],
  [lists:flatten(FormatedError) | format_errors(Left)];
format_errors([]) ->
  [].
  
time_now() ->
  T1 = hipe_bifs:get_hrvtime(),
  {time_now,T1}.

time_since({time_now,T1}) ->
  T2 = hipe_bifs:get_hrvtime(),
  T = T2-T1,
  trunc(T).

%%============ Get module names ==============

get_modules_by_dir(Files, StartFrom)->
  get_modules_by_dir(Files, StartFrom, []).

get_modules_by_dir([File|Left], StartFrom, Acc)->
  BaseName = 
    case StartFrom of
      src_code -> filename:basename(File, ".erl");
      byte_code -> filename:basename(File, ".beam")
    end,
  Dir = filename:dirname(File),
  get_modules_by_dir(Left, StartFrom, [{BaseName, Dir}|Acc]);
get_modules_by_dir([], _StartFrom, Acc) ->
  lists:reverse(Acc).
  

%%============ Send and receive functions ==============

send_log(Parent, Msg) ->
  Parent ! {self(), log, Msg}.

send_error(Parent, Msg) ->
  Parent ! {self(), error, Msg}.

send_all_warnings(Parent) ->
  case get_all_warnings() of
    [] -> ok;
    Warnings -> Parent ! {self(), warnings, Warnings}
  end.

get_all_warnings() ->
  receive
    {type_warning, TW} ->
      [TW | get_all_warnings()]
  after 0 ->
      []
  end.

send_module_callgraph(Parent, MCG) ->
  Parent ! {self(), module_callgraph, MCG}.

done(Parent) ->
  Parent ! {self(), done}.

get_mod_arg_types(Mod0) ->
  Mod = case Mod0 of
	  {M, _Dir} -> 
	    list_to_atom(M);
	  M when is_list(M) -> list_to_atom(M);
	  M when is_atom(M) -> M
	end,
  receive
    {arg_call_types, MFA = {Mod, _, _}, Types} -> 
      [{MFA, Types}|get_mod_arg_types(Mod)]
  after 0 ->
      []
  end.

flush() ->
  receive
    _ -> flush()
  after 0 ->
      ok
  end.
  
      

%%============ Handle dets ==============

mk_tmp_dets(Dets) ->
  TmpDets = Dets++"Tmp",
  case file:copy(Dets, TmpDets) of
    {error, Reason} ->
      Msg = io_lib:format("*****Aborting! Could not make a temp copy of"
			  " the lookup table:\n ~p\n", [Reason]),
      exit({error, Msg});
    {ok, _} ->
      TmpDets
  end.

get_module_info(Dets, Mod0) ->
  Mod = case Mod0 of
	  {M, _Dir} -> list_to_atom(M);
	  M when is_list(M) -> list_to_atom(M)
	end,  
  {ok, _} = dets:open_file(Dets, []),
  Info = dets:match_object(Dets, {{Mod, '_', '_'}, '_'}),
  OrdDict = orddict:from_list(Info),
  dets:close(Dets),
  OrdDict.

%%============ Handle the documentation table ==============

insert_doc_info(_, #analysis_state{doc_ets=none}) ->
  ok;
insert_doc_info({Mod, _Dir}, #analysis_state{dets=Dets, doc_ets=Ets}) ->
  ReturnTypes = get_module_info(Dets, Mod),
  Args = get_mod_arg_types(Mod),
  insert_doc_info_mfa_list(Args, ReturnTypes, Ets).

insert_doc_info_mfa_list([{MFA, ArgTypes}|Left], ReturnTypes, Ets) ->
  NewReturnTypes =
    case orddict:find(MFA, ReturnTypes) of
      {ok, Return} ->
	ets:insert(Ets, {MFA, {ArgTypes, Return}}),
	orddict:erase(MFA, ReturnTypes);
      error ->
	ets:insert(Ets, {MFA, {ArgTypes, erl_types:t_any()}}),
	ReturnTypes
    end,
  insert_doc_info_mfa_list(Left, NewReturnTypes, Ets);
insert_doc_info_mfa_list([], [{MFA, ReturnType}|Left], Ets) ->
  ets:insert(Ets, {MFA, ReturnType}),
  insert_doc_info_mfa_list([], Left, Ets);
insert_doc_info_mfa_list([], [], _Ets) ->
  ok.
